/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.transport.pipe.flow;

import java.util.EnumSet;
import java.util.function.Supplier;

import javax.annotation.Nonnull;

import net.minecraft.item.EnumDyeColor;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;

import buildcraft.lib.misc.NBTUtilBC;
import buildcraft.lib.misc.StackUtil;
import buildcraft.lib.misc.VecUtil;

public class TravellingItem {
    // Client fields - public for rendering
    @Nonnull
    public final Supplier<ItemStack> clientItemLink;
    public int stackSize;
    public EnumDyeColor colour;

    // Server fields
    /** The server itemstack */
    @Nonnull
    ItemStack stack;
    int id = 0;
    boolean toCenter;
    double speed = 0.05;
    /** Absolute times (relative to world.getTotalWorldTime()) with when an item started to when it finishes. */
    long tickStarted, tickFinished;
    /** Relative times (from tickStarted) until an event needs to be fired or this item needs changing. */
    int timeToDest;
    /** If {@link #toCenter} is true then this represents the side that the item is coming from, otherwise this
     * represents the side that the item is going to. */
    EnumFacing side;
    /** A set of all the faces that this item has tried to go and failed. */
    EnumSet<EnumFacing> tried = EnumSet.noneOf(EnumFacing.class);
    /** If true then events won't be fired for this, and this item won't be dropped by the pipe. However it will affect
     * pipe.isEmpty and related gate triggers. */
    boolean isPhantom = false;

    // @formatter:off
    /* States (server side):
      
      - TO_CENTER:
        - tickStarted is the tick that the item entered the pipe (or bounced back)
        - tickFinished is the tick that the item will reach the center 
        - side is the side that the item came from
        - timeToDest is equal to timeFinished - timeStarted
      
      - TO_EXIT:
       - tickStarted is the tick that the item reached the center
       - tickFinished is the tick that the item will reach the end of a pipe 
       - side is the side that the item is going to 
       - timeToDest is equal to timeFinished - timeStarted. 
     */
    // @formatter:on

    public TravellingItem(@Nonnull ItemStack stack) {
        this.stack = stack;
        clientItemLink = () -> ItemStack.field_190927_a;
    }

    public TravellingItem(Supplier<ItemStack> clientStackLink, int count) {
        this.clientItemLink = StackUtil.asNonNull(clientStackLink);
        this.stackSize = count;
        this.stack = StackUtil.EMPTY;
    }

    public TravellingItem(NBTTagCompound nbt, long tickNow) {
        clientItemLink = () -> ItemStack.field_190927_a;
        stack = new ItemStack(nbt.func_74775_l("stack"));
        int c = nbt.func_74771_c("colour");
        this.colour = c == 0 ? null : EnumDyeColor.func_176764_b(c - 1);
        this.toCenter = nbt.func_74767_n("toCenter");
        this.speed = nbt.func_74769_h("speed");
        if (speed < 0.001) {
            // Just to make sure that we don't have an invalid speed
            speed = 0.001;
        }
        tickStarted = nbt.func_74762_e("tickStarted") + tickNow;
        tickFinished = nbt.func_74762_e("tickFinished") + tickNow;
        timeToDest = nbt.func_74762_e("timeToDest");

        side = NBTUtilBC.readEnum(nbt.func_74781_a("side"), EnumFacing.class);
        if (side == null || timeToDest == 0) {
            // Older 8.0.x. version
            toCenter = true;
        }
        tried = NBTUtilBC.readEnumSet(nbt.func_74781_a("tried"), EnumFacing.class);
        isPhantom = nbt.func_74767_n("isPhantom");
    }

    public NBTTagCompound writeToNbt(long tickNow) {
        NBTTagCompound nbt = new NBTTagCompound();
        nbt.func_74782_a("stack", stack.serializeNBT());
        nbt.func_74774_a("colour", (byte) (colour == null ? 0 : colour.func_176765_a() + 1));
        nbt.func_74757_a("toCenter", toCenter);
        nbt.func_74780_a("speed", speed);
        nbt.func_74768_a("tickStarted", (int) (tickStarted - tickNow));
        nbt.func_74768_a("tickFinished", (int) (tickFinished - tickNow));
        nbt.func_74768_a("timeToDest", timeToDest);
        nbt.func_74782_a("side", NBTUtilBC.writeEnum(side));
        nbt.func_74782_a("tried", NBTUtilBC.writeEnumSet(tried, EnumFacing.class));
        if (isPhantom) {
            nbt.func_74757_a("isPhantom", true);
        }
        return nbt;
    }

    public int getCurrentDelay(long tickNow) {
        long diff = tickFinished - tickNow;
        if (diff < 0) {
            return 0;
        } else {
            return (int) diff;
        }
    }

    public double getWayThrough(long now) {
        long diff = tickFinished - tickStarted;
        long nowDiff = now - tickStarted;
        return nowDiff / (double) diff;
    }

    public void genTimings(long now, double distance) {
        tickStarted = now;
        timeToDest = (int) Math.ceil(distance / speed);
        tickFinished = now + timeToDest;
    }

    public boolean canMerge(TravellingItem with) {
        if (isPhantom || with.isPhantom) {
            return false;
        }
        return toCenter == with.toCenter//
            && colour == with.colour//
            && side == with.side//
            && Math.abs(tickFinished - with.tickFinished) < 4//
            && stack.func_77976_d() >= stack.func_190916_E() + with.stack.func_190916_E()//
            && StackUtil.canMerge(stack, with.stack);
    }

    /** Attempts to merge the two travelling item's together, if they are close enough.
     * 
     * @param with
     * @return */
    public boolean mergeWith(TravellingItem with) {
        if (canMerge(with)) {
            this.stack.func_190917_f(with.stack.func_190916_E());
            return true;
        }
        return false;
    }

    public Vec3d interpolatePosition(Vec3d start, Vec3d end, long tick, float partialTicks) {
        long diff = tickFinished - tickStarted;
        long nowDiff = tick - tickStarted;
        double sinceStart = nowDiff + partialTicks;
        double interpMul = sinceStart / diff;
        double oneMinus = 1 - interpMul;
        if (interpMul <= 0) return start;
        if (interpMul >= 1) return end;

        double x = oneMinus * start.field_72450_a + interpMul * end.field_72450_a;
        double y = oneMinus * start.field_72448_b + interpMul * end.field_72448_b;
        double z = oneMinus * start.field_72449_c + interpMul * end.field_72449_c;
        return new Vec3d(x, y, z);
    }

    public Vec3d getRenderPosition(BlockPos pos, long tick, float partialTicks, PipeFlowItems flow) {
        long diff = tickFinished - tickStarted;
        long afterTick = tick - tickStarted;

        float interp = (afterTick + partialTicks) / diff;
        interp = Math.max(0, Math.min(1, interp));

        Vec3d center = new Vec3d(pos).func_72441_c(0.5, 0.5, 0.5);
        Vec3d vecSide = side == null ? center : VecUtil.offset(center, side, flow.getPipeLength(side));

        Vec3d vecFrom;
        Vec3d vecTo;
        if (toCenter) {
            vecFrom = vecSide;
            vecTo = center;
        } else {
            vecFrom = center;
            vecTo = vecSide;
        }

        return VecUtil.scale(vecFrom, 1 - interp).func_178787_e(VecUtil.scale(vecTo, interp));
    }

    public EnumFacing getRenderDirection(long tick, float partialTicks) {
        long diff = tickFinished - tickStarted;
        long afterTick = tick - tickStarted;

        float interp = (afterTick + partialTicks) / diff;
        interp = Math.max(0, Math.min(1, interp));
        if (toCenter) {
            return side == null ? null : side.func_176734_d();
        } else {
            return side;
        }
    }

    public boolean isVisible() {
        return true;
    }
}
